"""
ddadevops provide tools to support builds combining gopass, 
terraform, dda-pallet, aws & hetzner-cloud.

"""

from .credential import gopass_credential_from_env_path, gopass_credential_from_path, gopass_password_from_path, gopass_field_from_path
from .devops_build import DevopsBuild, create_devops_build_config, get_devops_build
from .devops_terraform_build import WorkaroundTerraform, DevopsTerraformBuild, create_devops_terraform_build_config
from .devops_docker_build import DevopsDockerBuild, create_devops_docker_build_config
from .hetzner_mixin import HetznerMixin, add_hetzner_mixin_config
from .aws_backend_properties_mixin import AwsBackendPropertiesMixin, add_aws_backend_properties_mixin_config
from .aws_mfa_mixin import AwsMfaMixin, add_aws_mfa_mixin_config
from .aws_rds_pg_mixin import AwsRdsPgMixin, add_aws_rds_pg_mixin_config
from .dda_tenant_mixin import DdaTenantMixin, add_dda_tenant_mixin_config
from .dda_simple_mixin import DdaSimpleMixin, add_dda_simple_mixin_config

__version__ = "${version}"


from os import path
from json import load, dumps
from subprocess import run
from pkg_resources import *
from python_terraform import *
from .python_util import filter_none
from .devops_build import DevopsBuild, create_devops_build_config
import sys


def create_devops_terraform_build_config(stage,
                                         project_root_path,
                                         module,
                                         additional_vars,
                                         build_dir_name='target',
                                         output_json_name='output.json',
                                         use_workspace=True,
                                         use_package_common_files=True,
                                         build_commons_path=None,
                                         terraform_build_commons_dir_name='terraform',
                                         debug_print_terraform_command=False,
                                         additional_tfvar_files=[]):
    ret = create_devops_build_config(
        stage, project_root_path, module, build_dir_name)
    ret.update({'additional_vars': additional_vars,
                'output_json_name': output_json_name,
                'use_workspace': use_workspace,
                'use_package_common_files': use_package_common_files,
                'build_commons_path': build_commons_path,
                'terraform_build_commons_dir_name': terraform_build_commons_dir_name,
                'debug_print_terraform_command': debug_print_terraform_command,
                'additional_tfvar_files': additional_tfvar_files})
    return ret


class WorkaroundTerraform(Terraform):
    def __init__(self, working_dir=None,
                 targets=None,
                 state=None,
                 variables=None,
                 parallelism=None,
                 var_file=None,
                 terraform_bin_path=None,
                 is_env_vars_included=True,
                 ):
        super().__init__(working_dir, targets, state, variables, parallelism,
                         var_file,  terraform_bin_path, is_env_vars_included)
        self.latest_cmd = ''

    def generate_cmd_string(self, cmd, *args, **kwargs):
        result = super().generate_cmd_string(cmd, *args, **kwargs)
        self.latest_cmd = ' '.join(result)
        return result


class DevopsTerraformBuild(DevopsBuild):

    def __init__(self, project, config):
        super().__init__(project, config)
        project.build_depends_on('python-terraform')
        self.additional_vars = config['additional_vars']
        self.output_json_name = config['output_json_name']
        self.use_workspace = config['use_workspace']
        self.use_package_common_files = config['use_package_common_files']
        self.build_commons_path = config['build_commons_path']
        self.terraform_build_commons_dir_name = config['terraform_build_commons_dir_name']
        self.debug_print_terraform_command = config['debug_print_terraform_command']
        self.additional_tfvar_files = config['additional_tfvar_files']

    def terraform_build_commons_path(self):
        mylist = [self.build_commons_path,
                  self.terraform_build_commons_dir_name]
        return '/'.join(filter_none(mylist)) + '/'

    def project_vars(self):
        ret = {'stage': self.stage}
        if self.module:
            ret['module'] = self.module
        if self.additional_vars:
            ret.update(self.additional_vars)
        return ret

    def copy_build_resource_file_from_package(self, name):
        my_data = resource_string(
            __name__, "src/main/resources/terraform/" + name)
        with open(self.build_path() + '/' + name, "w") as output_file:
            output_file.write(my_data.decode(sys.stdout.encoding))

    def copy_build_resources_from_package(self):
        self.copy_build_resource_file_from_package('versions.tf')
        self.copy_build_resource_file_from_package('terraform_build_vars.tf')

    def copy_build_resources_from_dir(self):
        run('cp -f ' + self.terraform_build_commons_path() +
            '* ' + self.build_path(), shell=True)

    def copy_local_state(self):
        run('cp terraform.tfstate '  + self.build_path(), shell=True)

    def rescue_local_state(self):
        run('cp ' + self.build_path() + '/terraform.tfstate .', shell=True)

    def initialize_build_dir(self):
        super().initialize_build_dir()
        if self.use_package_common_files:
            self.copy_build_resources_from_package()
        else:
            self.copy_build_resources_from_dir()
        self.copy_local_state()
        run('cp *.tf ' + self.build_path(), shell=True)
        run('cp *.properties ' + self.build_path(), shell=True)       
        run('cp *.tfvars ' + self.build_path(), shell=True)
        run('cp -r scripts ' + self.build_path(), shell=True)

    def post_build(self):
        self.rescue_local_state()
        
    def init_client(self):
        tf = WorkaroundTerraform(working_dir=self.build_path())
        tf.init()
        self.print_terraform_command(tf)
        if self.use_workspace:
            try:
                tf.workspace('select', self.stage)
                self.print_terraform_command(tf)
            except:
                tf.workspace('new', self.stage)
                self.print_terraform_command(tf)
        return tf

    def write_output(self, tf):
        result = tf.output(json=IsFlagged)
        self.print_terraform_command(tf)
        with open(self.build_path() + self.output_json_name, "w") as output_file:
            output_file.write(json.dumps(result))

    def read_output_json(self):
        with open(self.build_path() + self.output_json_name, 'r') as f:
            return load(f)

    def plan(self):
        tf = self.init_client()
        tf.plan(capture_output=False, raise_on_error=False,
                var=self.project_vars(),
                var_file=self.additional_tfvar_files)
        self.post_build()
        self.print_terraform_command(tf)

    def plan_fail_on_diff(self):
        tf = self.init_client()
        tf.plan(detailed_exitcode=IsFlagged, capture_output=False, raise_on_error=False,
                var=self.project_vars(),
                var_file=self.additional_tfvar_files)
        self.post_build()
        self.print_terraform_command(tf)


    def apply(self, auto_approve=False):
        tf = self.init_client()
        tf.apply(capture_output=False, raise_on_error=True,
                 skip_plan=auto_approve,
                 var=self.project_vars(),
                 var_file=self.additional_tfvar_files)
        self.write_output(tf)
        self.post_build()
        self.print_terraform_command(tf)

    def destroy(self, auto_approve=False):
        tf = self.init_client()
        if auto_approve:
            force = IsFlagged
        else:
            force = None
        tf.destroy(capture_output=False, raise_on_error=True,
                   force=force,
                   var=self.project_vars(),
                   var_file=self.additional_tfvar_files)
        self.post_build()
        self.print_terraform_command(tf)

    def tf_import(self, tf_import_name, tf_import_resource,):
        tf = self.init_client()
        tf.import_cmd(tf_import_name, tf_import_resource,
                      capture_output=False, raise_on_error=True,
                      var=self.project_vars(),
                      var_file=self.additional_tfvar_files)
        self.post_build()
        self.print_terraform_command(tf)

    def print_terraform_command(self, tf):
        if self.debug_print_terraform_command:
            output = 'cd ' + self.build_path() + ' && ' + tf.latest_cmd
            print(output)



from python_terraform import *
from boto3 import *
from .python_util import execute
from .aws_backend_properties_mixin import AwsBackendPropertiesMixin


def add_aws_mfa_mixin_config(config, account_id, region,
                             mfa_role='developer', mfa_account_prefix='', 
                             mfa_login_account_suffix='main'):
    config.update({'AwsMfaMixin':
                   {'account_id': account_id,
                    'region': region,
                    'mfa_role': mfa_role,
                    'mfa_account_prefix': mfa_account_prefix,
                    'mfa_login_account_suffix': mfa_login_account_suffix}})
    return config


class AwsMfaMixin(AwsBackendPropertiesMixin):

    def __init__(self, project, config):
        super().__init__(project, config)
        project.build_depends_on('boto3')
        project.build_depends_on('mfa')
        aws_mfa_mixin_config = config['AwsMfaMixin']
        self.account_id = aws_mfa_mixin_config['account_id']
        self.region = aws_mfa_mixin_config['region']
        self.mfa_role = aws_mfa_mixin_config['mfa_role']
        self.mfa_account_prefix = aws_mfa_mixin_config['mfa_account_prefix']
        self.mfa_login_account_suffix = aws_mfa_mixin_config['mfa_login_account_suffix']

    def project_vars(self):
        ret = super().project_vars()
        ret.update({'account_name': self.account_name,
                    'account_id': self.account_id,
                    'region': self.region,
                    'mfa_role': self.mfa_role,
                    'mfa_account_prefix': self.mfa_account_prefix,
                    'mfa_login_account_suffix': self.mfa_login_account_suffix})
        return ret

    def get_username_from_account(self, p_account_name):
        login_id = execute('cat ~/.aws/accounts | grep -A 2 "\[' + p_account_name +
                           '\]"  | grep username | awk -F= \'{print $2}\'', shell=True)
        return login_id

    def get_account_id_from_account(self, p_account_name):
        account_id = execute('cat ~/.aws/accounts | grep -A 2 "\[' + p_account_name +
                             '\]"  | grep account | awk -F= \'{print $2}\'', shell=True)
        return account_id

    def get_mfa(self, mfa_path='aws'):
        mfa_token = execute('mfa otp ' + mfa_path, shell=True)
        return mfa_token

    def write_aws_config(self, to_profile, key, secret):
        execute('aws configure --profile ' + to_profile +
                ' set ' + key + ' ' + secret, shell=True)

    def get_mfa_session(self, toke=None):
        from_account_name = self.mfa_account_prefix + self.mfa_login_account_suffix
        from_account_id = self.get_account_id_from_account(from_account_name)
        to_account_name = self.mfa_account_prefix + self.account_name
        to_account_id = self.get_account_id_from_account(to_account_name)
        login_id = self.get_username_from_account(from_account_name)
        mfa_token = self.get_mfa()
        ses = Session(profile_name=from_account_name)
        sts_client = ses.client('sts')
        response = sts_client.assume_role(
            RoleArn='arn:aws:iam::' + to_account_id + ':role/' + self.mfa_role,
            RoleSessionName=to_account_id + '-' + self.account_name + '-' + self.mfa_role,
            SerialNumber='arn:aws:iam::' + from_account_id + ':mfa/' + login_id,
            TokenCode=mfa_token
        )

        self.write_aws_config(to_account_name, 'aws_access_key_id',
                              response['Credentials']['AccessKeyId'])
        self.write_aws_config(to_account_name, 'aws_secret_access_key',
                              response['Credentials']['SecretAccessKey'])
        self.write_aws_config(to_account_name, 'aws_session_token',
                              response['Credentials']['SessionToken'])
        print('got token')



from string import Template
from subprocess import run
from .python_util import *
from .devops_build import DevopsBuild


def add_dda_simple_mixin_config(config, domain_file_name,
                                jar_file=None,
                                target_edn_name='target.edn'):
    config.update({'DdaSimpleMixin':
                   {'domain_file_name': domain_file_name,
                    'target_edn_name': target_edn_name,
                    'jar_file': jar_file,
                    'target_template':
                    """
{:existing [{:node-name "$node_name"
             :node-ip "$ipv4"}]
 :provisioning-user {:login "root"}}
""", }})
    return config


class DdaSimpleMixin(DevopsBuild):

    def __init__(self, project, config):
        super().__init__(project, config)
        dda_pallet_mixin_config = config['DdaSimpleMixin']
        self.domain_file_name = dda_pallet_mixin_config['domain_file_name']
        self.target_edn_name = dda_pallet_mixin_config['target_edn_name']
        self.jar_file = dda_pallet_mixin_config['jar_file']
        if not self.jar_file:
            self.jar_file = 'target/uberjar/' + self.module + '-standalone.jar'
        self.target_template = Template(
            dda_pallet_mixin_config['target_template'])

    def initialize_build_dir(self):
        super().initialize_build_dir()
        run('cp *.edn ' + self.build_path(), shell=True)

    def dda_write_target(self, node_name, ipv4):
        with open(self.build_path() + '/' + self.target_edn_name, "w") as output_file:
            output_file.write(
                self.target_template.substitute({'ipv4': ipv4, 'node_name': node_name}))

    def dda_write_domain(self, substitues):
        with open(self.build_path() + '/' + self.domain_file_name, "r") as input_file:
            domain_input = input_file.read()
        domain_template = Template(domain_input)
        with open(self.build_path() + '/out_' + self.domain_file_name, "w") as output_file:
            output_file.write(domain_template.substitute(substitues))

    def dda_uberjar(self, configure_switch=None):
        if configure_switch:
            cmd = ['java', '-jar', self.project_root_path + '/' + self.jar_file,
                   '--targets', self.build_path() + '/' + self.target_edn_name,
                   '--configure',
                   self.build_path() + '/out_' + self.domain_file_name]
        else:
            cmd = ['java', '-jar', self.project_root_path + '/' + self.jar_file,
                   '--targets', self.build_path() + '/' + self.target_edn_name,
                   self.build_path() + '/out_' + self.domain_file_name]
        prn_cmd = list(cmd)
        print(" ".join(prn_cmd))
        output = execute(cmd)
        print(output)
        return output

    def dda_install(self):
        return self.dda_uberjar()

    def dda_configure(self):
        return self.dda_uberjar(True)



from subprocess import run
from .python_util import filter_none
from pkg_resources import *
from .devops_build import DevopsBuild, create_devops_build_config
import sys


def create_devops_docker_build_config(stage,
                                      project_root_path,
                                      module,
                                      dockerhub_user,
                                      dockerhub_password,
                                      build_dir_name='target',
                                      use_package_common_files=True,
                                      build_commons_path=None,
                                      docker_build_commons_dir_name='docker',):
    ret = create_devops_build_config(
        stage, project_root_path, module, build_dir_name)
    ret.update({'dockerhub_user': dockerhub_user,
                'dockerhub_password': dockerhub_password,
                'use_package_common_files': use_package_common_files,
                'docker_build_commons_dir_name': docker_build_commons_dir_name, 
                'build_commons_path': build_commons_path, })
    return ret


class DevopsDockerBuild(DevopsBuild):

    def __init__(self, project, config):
        super().__init__(project, config)
        project.build_depends_on('python-terraform')
        self.dockerhub_user = config['dockerhub_user']
        self.dockerhub_password = config['dockerhub_password']
        self.use_package_common_files = config['use_package_common_files']
        self.build_commons_path = config['build_commons_path']
        self.docker_build_commons_dir_name = config['docker_build_commons_dir_name']

    def docker_build_commons_path(self):
        mylist = [self.build_commons_path,
                  self.docker_build_commons_dir_name]
        return '/'.join(filter_none(mylist)) + '/'

    def copy_build_resource_file_from_package(self, name):
        run('mkdir -p ' + self.build_path() + '/image/resources', shell=True)
        my_data = resource_string(
            __name__, "src/main/resources/docker/" + name)
        with open(self.build_path() + '/' + name, "w") as output_file:
            output_file.write(my_data.decode(sys.stdout.encoding))

    def copy_build_resources_from_package(self):
        self.copy_build_resource_file_from_package(
            'image/resources/install_functions.sh')

    def copy_build_resources_from_dir(self):
        run('cp -f ' + self.docker_build_commons_path() +
            '* ' + self.build_path(), shell=True)

    def initialize_build_dir(self):
        super().initialize_build_dir()
        if self.use_package_common_files:
            self.copy_build_resources_from_package()
        else:
            self.copy_build_resources_from_dir()
        run('cp -r image ' + self.build_path(), shell=True)
        run('cp -r test ' + self.build_path(), shell=True)

    def image(self):
        run('docker build -t ' + self.name() +
            ' --file ' + self.build_path() + '/image/Dockerfile '
            + self.build_path() + '/image', shell=True)

    def drun(self):
        run('docker run --expose 8080 -it ' +
            self.name() + ' /bin/bash', shell=True)

    def dockerhub_login(self):
        run('docker login --username ' + self.dockerhub_user +
            ' --password ' + self.dockerhub_password, shell=True)

    def dockerhub_publish(self):
        run('docker tag ' + self.name() + ' ' + self.dockerhub_user +
            '/' + self.name(), shell=True)
        run('docker push ' + self.dockerhub_user +
            '/' + self.name(), shell=True)

    def test(self):
        run('docker build -t ' + self.name() + '-test ' +
            '--file ' + self.build_path() + '/test/Dockerfile '
            + self.build_path() + '/test', shell=True)



from python_terraform import *
from .devops_terraform_build import DevopsTerraformBuild, WorkaroundTerraform


def add_aws_backend_properties_mixin_config(config, account_name):
    config.update({'AwsBackendPropertiesMixin':
                   {'account_name': account_name}})
    return config


class AwsBackendPropertiesMixin(DevopsTerraformBuild):

    def __init__(self, project, config):
        super().__init__(project, config)
        aws_mixin_config = config['AwsBackendPropertiesMixin']
        self.account_name = aws_mixin_config['account_name']
        self.backend_config = "backend." + self.account_name + "." + self.stage + ".properties"
        self.additional_tfvar_files.append(self.backend_config)

    def project_vars(self):
        ret = super().project_vars()
        ret.update({'account_name': self.account_name})
        return ret

    def copy_build_resources_from_package(self):
        super().copy_build_resources_from_package()
        self.copy_build_resource_file_from_package('aws_provider.tf')
        self.copy_build_resource_file_from_package(
            'aws_backend_properties_vars.tf')
        self.copy_build_resource_file_from_package(
            'aws_backend_with_properties.tf')

    def copy_local_state(self):
        pass

    def rescue_local_state(self):
        pass

    def init_client(self):
        tf = WorkaroundTerraform(working_dir=self.build_path())
        tf.init(backend_config=self.backend_config)
        self.print_terraform_command(tf)
        if self.use_workspace:
            try:
                tf.workspace('select', slef.stage)
                self.print_terraform_command(tf)
            except:
                tf.workspace('new', self.stage)
                self.print_terraform_command(tf)
        return tf



from .python_util import *
import deprecation

@deprecation.deprecated(deprecated_in="0.5.0", removed_in="1.0",
                        details='use gopass_password_from_path(os.environ.get(env_var, None)) instead')
def gopass_credential_from_env_path (env_var):
    env_path = os.environ.get(env_var, None)
    return gopass_password_from_path(env_path)

@deprecation.deprecated(deprecated_in="0.5.0", removed_in="1.0",
                        details='use gopass_password_from_path(path) instead')
def gopass_credential_from_path (path):
    return gopass_password_from_path(path)


def gopass_field_from_path (path, field):
    credential = None
    if path and field:
        print('get field for: ' + path + ', ' + field)
        credential = execute(['gopass', 'show', path, field])
    return credential

def gopass_password_from_path (path):
    credential = None
    if path:
        print('get password for: ' + path)
        credential = execute(['gopass', 'show', '--password', path])
    return credential




#!/usr/bin/env python
#   -*- coding: utf-8 -*-

from setuptools import setup
from setuptools.command.install import install as _install

class install(_install):
    def pre_install_script(self):
        pass

    def post_install_script(self):
        pass

    def run(self):
        self.pre_install_script()

        _install.run(self)

        self.post_install_script()

if __name__ == '__main__':
    setup(
        name = 'ddadevops',
        version = '0.8.11.dev20200703135042',
        description = 'tools to support builds combining gopass, terraform, dda-pallet, aws & hetzner-cloud',
        long_description = '# dda-devops-build\n\n[![Slack](https://img.shields.io/badge/chat-clojurians-green.svg?style=flat)](https://clojurians.slack.com/messages/#dda-pallet/) | [<img src="https://meissa-gmbh.de/img/community/Mastodon_Logotype.svg" width=20 alt="team@social.meissa-gmbh.de"> team@social.meissa-gmbh.de](https://social.meissa-gmbh.de/@team) | [Website & Blog](https://domaindrivenarchitecture.org)\n\n![release prod](https://github.com/DomainDrivenArchitecture/dda-devops-build/workflows/release%20prod/badge.svg)\n\ndda-devops-build provide a envioronment to tie several DevOps tools together for easy interoperation. Supported tools are:\n* aws with\n  * simple api-key auth\n  * mfa & assume-role auth\n* hetzner with simple api-key auth\n* terraform v0.11, v0.12 supporting\n  * local file backends\n  * s3 backends\n* docker / dockerhub\n* user / team credentials managed by gopass\n* dda-pallet\n\n# Setup\n\n```\nsudo apt install python3-pip\npip3 install pip3 --upgrade --user\npip3 install pybuilder ddadevops deprecation --user\nexport PATH=$PATH:~/.local/bin\n\n# in case of using terraform\npip3 install python-terraform --user\n\n# in case of using AwsMixin\npip3 install boto3 --user\n\n# in case of using AwsMfaMixin\npip3 install boto3 mfa --user\n```\n\n# Example Build\n\nlets assume the following poject structure\n\n```\nmy-project\n   | -> my-module\n   |       | -> build.py\n   |       | -> some-terraform.tf\n   | -> an-other-module\n   | -> target  (here will the build happen)\n   |       | -> ...\n```\n\n```\nfrom pybuilder.core import task, init\nfrom ddadevops import *\n\nname = \'my-project\'\nMODULE = \'my-module\'\nPROJECT_ROOT_PATH = \'..\'\n\nclass MyBuild(DevopsTerraformBuild):\n    pass\n\n\n@init\ndef initialize(project):\n    project.build_depends_on(\'ddadevops>=0.5.0\')\n    account_name = \'my-aws-account-name\'\n    account_id = \'my-aws-account-id\'\n    stage = \'my stage i.e. dev|test|prod\'\n    additional_vars = {\'var_to_use_insied_terraform\': \'...\'}\n    additional_var_files = [\'variable-\' + account_name + \'-\' + stage + \'.tfvars\']\n    config = create_devops_terraform_build_config(stage, PROJECT_ROOT_PATH,\n                                                  MODULE, additional_vars,\n                                                  additional_tfvar_files=additional_var_files)\n    build = MyBuild(project, config)\n    build.initialize_build_dir()\n\n\n@task\ndef plan(project):\n    build = get_devops_build(project)\n    build.plan()\n\n\n@task\ndef apply(project):\n    build = get_devops_build(project)\n    build.apply()\n\n@task\ndef destroy(project):\n    build = get_devops_build(project)\n    build.destroy()\n\n@task\ndef tf_import(project):\n    build = get_devops_build(project)\n    build.tf_import(\'aws_resource.choosen_name\', \'the_aws_id\')\n```\n\n## Feature aws-backend\n\nWill use a file `backend.dev.live.properties` where dev is the [account-name], live is the  [stage].\n\nthe backend.dev.live.properties file content:\n```\nkey = ".."\nregion = "the aws region"\nprofile = "the profile used for aws"\nbucket = "the s3 bucket name"\nkms_key_id = "the aws key id"\n```\n\nthe build.py file content:\n```\nclass MyBuild(AwsBackendPropertiesMixin, DevopsTerraformBuild):\n    pass\n\n\n@init\ndef initialize(project):\n    project.build_depends_on(\'ddadevops>=0.5.0\')\n    account_name = \'my-aws-account-name\'\n    account_id = \'my-aws-account-id\'\n    stage = \'my stage i.e. dev|test|prod\'\n    additional_vars = {}\n    config = create_devops_terraform_build_config(stage, PROJECT_ROOT_PATH,\n                                                  MODULE, additional_vars)\n    config = add_aws_backend_properties_mixin_config(config, account_name)\n    build = MyBuild(project, config)\n    build.initialize_build_dir()\n```\n\n## Feature aws-mfa-assume-role\n\nIn order to use aws assume role in combination with the mfa-tool (`pip install mfa`):\n\nthe build.py file content:\n```\nclass MyBuild(class MyBuild(AwsMfaMixin, DevopsTerraformBuild):\n    pass\n\n\n@init\ndef initialize(project):\n    project.build_depends_on(\'ddadevops>=0.5.0\')\n    account_name = \'my-aws-account-name\'\n    account_id = \'my-aws-account-id\'\n    stage = \'my stage i.e. dev|test|prod\'\n    additional_vars = {}\n    config = create_devops_terraform_build_config(stage, PROJECT_ROOT_PATH,\n                                                  MODULE, additional_vars)\n    config = add_aws_backend_properties_mixin_config(config, account_name)\n    config = add_aws_mfa_mixin_config(config, account_id, \'eu-central-1\',\n                                      mfa_role=\'my_developer_role\',\n                                      mfa_account_prefix=\'company-\',\n                                      mfa_login_account_suffix=\'users_are_defined_here\')\n    build = MyBuild(project, config)\n    build.initialize_build_dir()\n\n@task\ndef access(project):\n    build = get_devops_build(project)\n    build.get_mfa_session()\n```\n\n## Feature DdaDockerBuild\n\nThe docker build supports image building, tagging, testing and login to dockerhost.\nFor bash based builds we support often used script-parts as predefined functions [see install_functions.sh](src/main/resources/docker/image/resources/install_functions.sh).\n\nA full working example: [doc/example/50_docker_module](doc/example/50_docker_module)\n\n## Feature AwsRdsPgMixin\n\nThe AwsRdsPgMixin provides \n* execute_pg_rds_sql - function will optionally resolve dns-c-names for trusted ssl-handshakes\n* alter_db_user_password\n* add_new_user\n* deactivate_user\n\nthe build.py file content:\n```\nclass MyBuild(..., AwsRdsPgMixin):\n    pass\n\n\n@init\ndef initialize(project):\n    project.build_depends_on(\'ddadevops>=0.8.0\')\n\n    ...\n    config = add_aws_rds_pg_mixin_config(config,\n                                         stage + "-db.bcsimport.kauf." + account_name + ".breuni.de",\n                                         "kauf_bcsimport",\n                                         rds_resolve_dns=True,)\n    build = MyBuild(project, config)\n    build.initialize_build_dir()\n\n@task\ndef rotate_credentials_in(project):\n    build = get_devops_build(project)\n    build.alter_db_user_password(\'/postgres/support\')\n    build.alter_db_user_password(\'/postgres/superuser\')\n    build.add_new_user(\'/postgres/superuser\', \'/postgres/app\', \'pg_group_role\')\n\n\n@task\ndef rotate_credentials_out(project):\n    build = get_devops_build(project)\n    build.deactivate_user(\'/postgres/superuser\', \'old_user_name\')\n```\n\n# Releasing and updating\n## Publish snapshot\n\n1. every push will be published as dev-dependency\n\n## Release\n1. git commit -am "release"\n1. git tag -am "release" version\n2. git push --follow-tags\n3. increase version no in build.py\n4. git commit -am "version bump"\n5. git push\n6. pip3 install --upgrade --user ddadevops\n\n\n# License\n\nCopyright © 2019 meissa GmbH\nLicensed under the [Apache License, Version 2.0](LICENSE) (the "License")\n',
        long_description_content_type = 'text/markdown',
        classifiers = [
            'License :: OSI Approved :: Apache Software License',
            'Programming Language :: Python',
            'Programming Language :: Python :: 2.7',
            'Programming Language :: Python :: 3',
            'Programming Language :: Python :: 3.5',
            'Programming Language :: Python :: 3.6',
            'Programming Language :: Python :: 3.7',
            'Programming Language :: Python :: 3.8',
            'Operating System :: POSIX :: Linux',
            'Operating System :: OS Independent',
            'Development Status :: 5 - Production/Stable',
            'Environment :: Console',
            'Intended Audience :: Developers',
            'License :: OSI Approved :: Apache Software License',
            'Topic :: Software Development :: Build Tools',
            'Topic :: Software Development :: Quality Assurance',
            'Topic :: Software Development :: Testing'
        ],
        keywords = '',

        author = 'meissa GmbH',
        author_email = 'buero@meissa-gmbh.de',
        maintainer = '',
        maintainer_email = '',

        license = 'Apache Software License',

        url = 'https://github.com/DomainDrivenArchitecture/dda-devops-build',
        project_urls = {},

        scripts = [],
        packages = ['ddadevops'],
        namespace_packages = [],
        py_modules = [],
        entry_points = {},
        data_files = [],
        package_data = {
            'ddadevops': ['LICENSE', 'src/main/resources/terraform/*', 'src/main/resources/docker/image/resources/*']
        },
        install_requires = [],
        dependency_links = [],
        zip_safe = True,
        cmdclass = {'install': install},
        python_requires = '!=3.0,!=3.1,!=3.2,!=3.3,!=3.4,<3.9,>=2.7',
        obsoletes = [],
    )



from .devops_terraform_build import DevopsTerraformBuild


def add_hetzner_mixin_config(config, hetzner_api_key):
    config.update({'HetznerMixin':
                   {'hetzner_api_key': hetzner_api_key}})
    return config


class HetznerMixin(DevopsTerraformBuild):

    def __init__(self, project, config):
        super().__init__(project, config)
        hetzner_mixin_config = config['HetznerMixin']
        self.hetzner_api_key = hetzner_mixin_config['hetzner_api_key']

    def project_vars(self):
        ret = super().project_vars()
        if self.hetzner_api_key:
            ret['hetzner_api_key'] = self.hetzner_api_key
        return ret

    def copy_build_resources_from_package(self):
        super().copy_build_resources_from_package()
        self.copy_build_resource_file_from_package('hetzner_provider.tf')
        self.copy_build_resource_file_from_package('hetzner_mixin_vars.tf')



from .python_util import execute
from .credential import gopass_password_from_path, gopass_field_from_path
from .devops_build import DevopsBuild


def add_aws_rds_pg_mixin_config(config, rds_host_name, db_name,
                                rds_resolve_dns=False,
                                db_port='5432'):
    config.update({'AwsRdsPgMixin':
                   {'rds_host_name': rds_host_name,
                    'db_name': db_name,
                    'rds_resolve_dns': rds_resolve_dns,
                    'db_port': db_port,
                    }})
    return config


class AwsRdsPgMixin(DevopsBuild):

    def __init__(self, project, config):
        super().__init__(project, config)
        aws_rds_pg_mixin_config = config['AwsRdsPgMixin']
        self.rds_host_name = aws_rds_pg_mixin_config['rds_host_name']
        self.rds_resolve_dns = aws_rds_pg_mixin_config['rds_resolve_dns']
        self.db_name = aws_rds_pg_mixin_config['db_name']
        self.db_port = aws_rds_pg_mixin_config['db_port']

    def execute_pg_rds_sql(self, user, password, sql):
        if self.rds_resolve_dns:
            host_cmd = "dig " + self.rds_host_name + " +short | head -n1"
            host = execute(host_cmd, shell=True)
        else:
            host = self.rds_host_name

        cmd = "PGUSER=" + user + " PGPASSWORD=" + password + \
            " psql --dbname=" + self.db_name + " --host=" + host + " --port=" + self.db_port + \
            " --set=sslmode=require -Atc \"" + sql + "\""
        result = execute(cmd, shell=True)
        print("PSQL: ", host, result.rstrip())
        return result

    def alter_db_user_password(self, gopass_path):
        user_name = gopass_field_from_path(gopass_path, 'user')
        user_old_password = gopass_field_from_path(gopass_path, 'old-password')
        user_new_password = gopass_password_from_path(gopass_path)

        self.execute_pg_rds_sql(user_name, user_old_password,
                                "ALTER ROLE " + user_name + " WITH PASSWORD '" + user_new_password + "';")
        print("changed password:", user_name)

    def add_new_user(self, gopass_path_superuser, gopass_path_new_user, group_role):
        superuser_name = gopass_field_from_path(gopass_path_superuser, 'user')
        superuser_password = gopass_password_from_path(gopass_path_superuser)
        new_user_name = gopass_field_from_path(gopass_path_new_user, 'user')
        new_user_password = gopass_password_from_path(gopass_path_new_user)

        self.execute_pg_rds_sql(superuser_name, superuser_password,
                                "CREATE ROLE " + new_user_name + " WITH LOGIN INHERIT PASSWORD '" + new_user_password + "';" +
                                "GRANT " + group_role + " TO " + new_user_name + ";")
        print("created user:", new_user_name)

    def deactivate_user(self, gopass_path_superuser, to_deactivate_user_name):
        superuser_name = gopass_field_from_path(gopass_path_superuser, 'user')
        superuser_password = gopass_password_from_path(gopass_path_superuser)

        owned_by_wrong_user = self.execute_pg_rds_sql(superuser_name, superuser_password,
                                                      "SELECT count(*) FROM pg_class c, pg_user u WHERE c.relowner = u.usesysid " +
                                                      "and u.usename='" + to_deactivate_user_name + "';")
        
        if int(owned_by_wrong_user) > 0:
            owned_objects = self.execute_pg_rds_sql(superuser_name, superuser_password,
                                                "SELECT c.relname FROM pg_class c, pg_user u WHERE c.relowner = u.usesysid " +
                                                "and u.usename='" + to_deactivate_user_name + "';")
            raise AssertionError(
                "There are still objects owned by the user to be deactivated:", owned_objects,to_deactivate_user_name)

        connections = self.execute_pg_rds_sql(superuser_name, superuser_password,
                                              "SELECT count(*) FROM pg_stat_activity WHERE application_name = " +
                                              "'PostgreSQL JDBC Driver' AND usename = '" + to_deactivate_user_name + "';")
        if int(connections) > 0:
            raise AssertionError("User is still connected.")

        self.execute_pg_rds_sql(superuser_name, superuser_password,
                                "ALTER ROLE " + to_deactivate_user_name + " WITH NOLOGIN NOCREATEROLE;")
        print('deactivated user:', to_deactivate_user_name)

    def change_owned_objects(self, gopass_path_superuser, to_deactivate_user_name, owner):
        superuser_name = gopass_field_from_path(gopass_path_superuser, 'user')
        superuser_password = gopass_password_from_path(gopass_path_superuser)

        alter_objects = f"""SELECT 'ALTER TABLE ' || c.relname || ' OWNER TO {owner};'
            FROM pg_class c, pg_user u
            WHERE c.relowner = u.usesysid 
            and c.relkind = 'r'
            and u.usename='{to_deactivate_user_name}'
            UNION
            SELECT 'ALTER INDEX ' || c.relname || ' OWNER TO {owner};'
            FROM pg_class c, pg_user u
            WHERE c.relowner = u.usesysid 
            and c.relkind = 'i'
            and c.relname not like 'pg_toast%'
            and u.usename='{to_deactivate_user_name}'
            UNION
            SELECT 'ALTER SEQUENCE ' || c.relname || ' OWNER TO {owner};'
            FROM pg_class c, pg_user u
            WHERE c.relowner = u.usesysid 
            and c.relkind = 'S'
            and u.usename='{to_deactivate_user_name}';"""

        alter_stmt = self.execute_pg_rds_sql(superuser_name, superuser_password, alter_objects)
        alter_stmt.strip()

        if (alter_stmt != ''):
            print('apply alter statements? \n', alter_stmt)
            proceed = input('\n[y/n] \n')
            if(proceed == 'y'):
                self.execute_pg_rds_sql(superuser_name, superuser_password, alter_stmt)
